"""Key functions for memoizing decorators."""
from typing import Callable, Type


HASH_CONFIG = {}


def register_hash_function(obj_type: Type, function: Callable):
    """Registers a user defined hash function for a given type"""
    HASH_CONFIG[obj_type] = function


try:
    import numpy as np

    register_hash_function(np.ndarray, lambda x: hash(str(x)))
except ImportError:
    pass


def _hash(obj):
    try:
        return hash(obj)
    except TypeError as exc:
        hasher = HASH_CONFIG.get(type(obj))
        if hasher:
            return hasher(obj)
        raise exc


def hash_params(func, *args, **kwargs):
    if kwargs is None:
        kwargs = {}

    code = func.__code__

    params = [code.co_name, code.co_filename, code.co_code] + list(args) + sorted(kwargs.items())

    hashes = []
    for param in params:
        param_hash = _hash(param)
        hashes.append(param_hash)
    return f'{code.co_name}:{code.co_filename}:{hash(tuple(hashes))}'



"""Extensible memoizing collections and decorators."""

from __future__ import absolute_import

from functools import update_wrapper

from . import keys
from .keys import hash_params, register_hash_function
from .cache import Cache
from .lfu import LFUCache
from .lru import LRUCache
from .rr import RRCache
from .sqlite3 import SQLite3Cache
from .ttl import TTLCache

__all__ = (
    "Cache",
    "LFUCache",
    "LRUCache",
    "RRCache",
    "SQLite3Cache",
    "TTLCache",
    "cached",
    "cachedmethod",
)

__project__ = "eidetic"
__version__ = "2018.12.2"


def cached(cache, key=hash_params, lock=None):
    """Decorator to wrap a function with a memoizing callable that saves
    results in a cache.
    """

    def decorator(func):

        if cache is None:

            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)

        elif lock is None:

            def wrapper(*args, **kwargs):
                k = key(func, *args, **kwargs)
                try:
                    return cache[k]
                except KeyError:
                    pass  # key not found
                v = func(*args, **kwargs)
                try:
                    cache[k] = v
                except ValueError:
                    pass  # value too large
                return v

        else:

            def wrapper(*args, **kwargs):
                k = key(func, *args, **kwargs)
                try:
                    with lock:
                        return cache[k]
                except KeyError:
                    pass  # key not found
                v = func(*args, **kwargs)
                try:
                    with lock:
                        cache[k] = v
                except ValueError:
                    pass  # value too large
                return v

        return update_wrapper(wrapper, func)

    return decorator


def cachedmethod(cache, key=hash_params, lock=None):
    """Decorator to wrap a class or instance method with a memoizing
    callable that saves results in a cache.

    """

    def decorator(method):
        if lock is None:

            def wrapper(self, *args, **kwargs):
                c = cache(self)
                if c is None:
                    return method(self, *args, **kwargs)
                k = key(*args, **kwargs)
                try:
                    return c[k]
                except KeyError:
                    pass  # key not found
                v = method(self, *args, **kwargs)
                try:
                    c[k] = v
                except ValueError:
                    pass  # value too large
                return v

        else:

            def wrapper(self, *args, **kwargs):
                c = cache(self)
                if c is None:
                    return method(self, *args, **kwargs)
                k = key(*args, **kwargs)
                try:
                    with lock(self):
                        return c[k]
                except KeyError:
                    pass  # key not found
                v = method(self, *args, **kwargs)
                try:
                    with lock(self):
                        c[k] = v
                except ValueError:
                    pass  # value too large
                return v

        return update_wrapper(wrapper, method)

    return decorator



from __future__ import absolute_import

import datetime as dt
import os
import pickle
import sqlite3
import zlib

from appdirs import user_cache_dir

from .abc import DefaultMapping


class _DefaultSize(object):
    def __getitem__(self, _):
        return 1

    def __setitem__(self, _, value):
        assert value == 1

    def pop(self, _):
        return 1


class SQLite3Cache(DefaultMapping):
    """Mutable mapping to serve as a simple cache or cache base class."""

    __size = _DefaultSize()

    def __init__(self, maxsize, getsizeof=None, app_name="eidetic", version="0.1"):
        self.cache_dir = user_cache_dir(appname=app_name, version=version)
        self.cache_path = self.cache_dir + os.sep + "eidetic.db"

        if not os.path.exists(self.cache_dir):
            os.makedirs(self.cache_dir)

        self.con = sqlite3.connect(self.cache_path)
        self.c = self.con.cursor()

        self.c.execute(
            """CREATE TABLE IF NOT EXISTS cache
                          (key text primary key,
                          val blob not null,
                          timestamp datetime default current_timestamp
                          );"""
        )

        if getsizeof:
            self.getsizeof = getsizeof
        if self.getsizeof is not SQLite3Cache.getsizeof:
            self.__size = dict()
        self.__data = dict()
        self.__currsize = 0
        self.__maxsize = maxsize

    def __repr__(self):
        return "%s(%r, maxsize=%r, currsize=%r)" % (
            self.__class__.__name__,
            list(self.__data.items()),
            self.__maxsize,
            self.__currsize,
        )

    def __getitem__(self, key):
        try:
            return self.__data[key]
        except KeyError:
            # Check if it exists in the databse
            bresult = self.c.execute(
                "SELECT val from cache where key = ?", (key,)
            ).fetchone()
            if bresult is not None:
                decomp_result = zlib.decompress(bresult[0])
                pyresult = pickle.loads(decomp_result)
                self.__data[key] = pyresult
                return pyresult
            return self.__missing__(key)

    def __setitem__(self, key, value):
        maxsize = self.__maxsize
        size = self.getsizeof(value)
        if size > maxsize:
            raise ValueError("value too large")
        if key not in self.__data or self.__size[key] < size:
            while self.__currsize + size > maxsize:
                self.popitem()
        if key in self.__data:
            diffsize = size - self.__size[key]
        else:
            diffsize = size
        self.c.execute(
            "INSERT OR REPLACE INTO cache VALUES (?, ?, ?)",
            (key, zlib.compress(pickle.dumps(value), -1), dt.datetime.utcnow()),
        )
        self.con.commit()
        self.__data[key] = value
        self.__size[key] = size
        self.__currsize += diffsize

    def __delitem__(self, key):
        size = self.__size.pop(key)
        self.c.execute("DELETE FROM cache WHERE key = ?", (key,))
        self.con.commit()
        del self.__data[key]
        self.__currsize -= size

    def __contains__(self, key):
        return key in self.__data

    def __missing__(self, key):
        raise KeyError(key)

    def __iter__(self):
        return iter(self.__data)

    def __len__(self):
        return len(self.__data)

    @property
    def maxsize(self):
        """The maximum size of the cache."""
        return self.__maxsize

    @property
    def currsize(self):
        """The current size of the cache."""
        return self.__currsize

    @staticmethod
    def getsizeof(value):
        """Return the size of a cache element's value."""
        return 1
